// XCamera.cpp: implementation of the XCamera class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#define		DISPLAY_API	__declspec(dllexport)

#include "../common/common.h"
#include "interfaces.h"
#include "manager.h"
#include "XCamera.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

XCamera::XCamera()
{
	m_lpDevice = NULL;
	m_lpCameraInfo = NULL;

	m_fZoomDistance = 50.0f;	

	m_vLookAtPoint = D3DXVECTOR3(10, 10, 10); // origin	
	m_fViewAngle = -D3DX_PI * 3 / 4; // 135 degree
	
	m_fScrlVelX = 0.0f;
	m_fScrlVelY = 0.0f;
	m_fZAngleVel = 0.0f;
	m_fRotVel = 0.0f;
	m_fZoomVel = 0.0f;

	m_iMinPos = 0;
	m_iMaxPos = 10;
	m_fZAngle = Z_ANGLE;
	m_dwRef = 0;

	m_fCameraHeight = 0.0f;	
	
	m_ePathStatus = STOP;
	m_fSpeed = 1.0f;
	m_bPathLoaded = FALSE;
	m_arPath = NULL;

	m_fTotalFPS = 0.0;
	m_fOldTime = 0.0;
	m_fAvgFPS = 0.0f;

	m_dwFlags = CF_CATMULLROMINT|CF_DRAWINFO;
}

XCamera::~XCamera()
{
	SAFE_RELEASE(m_lpCameraInfo);
	m_lpDevice = NULL;
	PathSegment *pNext = NULL;
	PathSegment *pToDelete = m_arPath;
	// Release path
	if(m_bPathLoaded)
	{
		while(pToDelete)
		{
			pNext = pToDelete->pNext;
			SAFE_DELETE(pToDelete);
			pToDelete = pNext;
			if(pNext == m_arPath)
			{
				break;
			}
		}
	}
}

HRESULT		XCamera::AddRef()
{
	return m_dwRef++;
}

HRESULT		XCamera::Release()
{
	if(m_dwRef > 0)
	{
		m_dwRef--;
	}
	if(m_dwRef == 0)
	{			
		delete this;
		TRACE("XCamera object destroyed.");
		return 0;
	}	
	else
	{
		return m_dwRef;
	}	
}
//
// Inicialization
HRESULT		XCamera::InitCamera(int iMinBound, int iMaxBound)
{
	HRESULT dwRet = S_FALSE;
	IDisplay *dis = NULL;	
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("InitCamera: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	SAFE_RELEASE(dis);
	if(m_lpDevice)
	{		
		m_iMinPos = iMinBound;
		m_iMaxPos = iMaxBound; 

		if(S_OK == CreateDisplayObject(DISIID_I3DFont, (void**)&m_lpCameraInfo))
		{
			m_lpCameraInfo->Load(BL_BITMAP, "fps.tga");		
			m_lpCameraInfo->Load(BL_SIZES, "fps.dat");
		}
		else
		{
			XException exp("InitCamera: Cannot create font for camera info.");
			THROW(exp);
		}
		//
		// Set transformations	
		D3DXMATRIX matWorld;		
		//
		// Set world matrix
		D3DXMatrixIdentity(&matWorld);
		m_lpDevice->SetTransform(D3DTS_WORLD, &matWorld);	
		//
		// Create view matrix
		D3DXVECTOR3	vUpDir;
		m_vEyePt.x = m_fZoomDistance * cosf(m_fViewAngle) + m_vLookAtPoint.x;
		m_vEyePt.y = m_fZoomDistance * sinf(m_fViewAngle) + m_vLookAtPoint.y;
		m_vEyePt.z = m_fZoomDistance * sinf(Z_ANGLE);
		//
		vUpDir = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
		//
		D3DXMatrixLookAtLH(&m_matView, &m_vEyePt, &m_vLookAtPoint, &vUpDir);	
		D3DXMatrixPerspectiveFovLH(&m_matProjection, FOV, ASPECT, NEAR_CLIP, FAR_CLIP);	
		
		m_lpDevice->SetTransform(D3DTS_VIEW, &m_matView);
		m_lpDevice->SetTransform(D3DTS_PROJECTION, &m_matProjection);		
		dwRet = S_OK;
	}
	return dwRet;
}

// Restores camera matrices
HRESULT		XCamera::RestoreCamera()
{
	D3DXMATRIX matWorld;		
	if(m_lpDevice)
	{	
		//
		// Set world matrix
		D3DXMatrixIdentity(&matWorld);
		m_lpDevice->SetTransform(D3DTS_WORLD, &matWorld);
		m_lpDevice->SetTransform(D3DTS_VIEW, &m_matView);
		m_lpDevice->SetTransform(D3DTS_PROJECTION, &m_matProjection);		
		return S_OK;
	}
	return S_FALSE;
}


// Update
HRESULT		XCamera::ProcessCamera(float	fElapsedTime)
{
	DWORD dwRet = S_FALSE; // CAMERA NOT INITIALIZED
	IDisplay *dis;
	float s = 0.0f, fDif;	
	if(m_lpDevice) {		
	
		D3DXVECTOR3	vUpDir;
		vUpDir = D3DXVECTOR3(0.0f, 0.0f, 1.0f);

		// Interactive mode 
		if(!(m_dwFlags & CF_PATH_MODE))
		{
			//
			// Apply rotation
			m_fViewAngle += m_fRotVel * fElapsedTime;
			//
			// Apply zooming
			m_fZoomDistance += m_fZoomVel * fElapsedTime;
			//
			// Check zoom distance
			if(m_fZoomDistance > MAX_ZOOM)
				m_fZoomDistance = MAX_ZOOM;
			if(m_fZoomDistance < MIN_ZOOM)
				m_fZoomDistance = MIN_ZOOM;
			//
			// Apply Z angle
			m_fZAngle += m_fZAngleVel * fElapsedTime;		
			if(m_fZAngle > MAX_Z_ANGLE)
				m_fZAngle = MAX_Z_ANGLE;
			if(m_fZAngle < MIN_Z_ANGLE)
				m_fZAngle = MIN_Z_ANGLE;
			//	
			// Apply position
			m_vLookAtPoint.x += (m_fScrlVelX * cosf(m_fViewAngle + D3DX_PI / 2)) * fElapsedTime;
			m_vLookAtPoint.y += (m_fScrlVelX * sinf(m_fViewAngle + D3DX_PI / 2)) * fElapsedTime;
			m_vLookAtPoint.x += (m_fScrlVelY * cosf(m_fViewAngle)) * fElapsedTime;
			m_vLookAtPoint.y += (m_fScrlVelY * sinf(m_fViewAngle)) * fElapsedTime;
			m_vLookAtPoint.z = m_fCameraHeight;
			//
			// Bound camera position
			if(m_vLookAtPoint.x < m_iMinPos) 
				m_vLookAtPoint.x = (float)m_iMinPos;
			if(m_vLookAtPoint.y < m_iMinPos) 
				m_vLookAtPoint.y = (float)m_iMinPos;
			if(m_vLookAtPoint.x > m_iMaxPos) 
				m_vLookAtPoint.x = (float)m_iMaxPos;
			if(m_vLookAtPoint.y > m_iMaxPos) 
				m_vLookAtPoint.y = (float)m_iMaxPos;

			//
			// Computes eye point position
			m_vEyePt.x = m_fZoomDistance * cosf(m_fViewAngle) + m_vLookAtPoint.x;
			m_vEyePt.y = m_fZoomDistance * sinf(m_fViewAngle) + m_vLookAtPoint.y;
			m_vEyePt.z = m_fZoomDistance * sinf(m_fZAngle) + m_fCameraHeight;
		}
		else
		{
			// Path mode 	
			PathSegment *ps1, *ps2, *ps3, *ps4;

			if(m_ePathStatus == PLAY)
			{
				fElapsedTime *= m_fSpeed;				
				fDif = m_curSeg->pNext->fTime - m_curSeg->fTime;							
				s = (m_curInterpolation.fTime - m_curSeg->fTime) / abs(fDif);
				// Next segment or last segment
				if((s > 1.0 || m_curSeg == m_arPath->pPrevious) && m_fSpeed > 0.0f) 
				{
					// Next segment
					if(m_curSeg->pNext->pNext)
					{
						m_curSeg = m_curSeg->pNext;							
					}
					// First segment
					if(m_curSeg == m_arPath)
					{
						m_fAvgFPS = m_fTotalFPS / m_arPath->pPrevious->fTime;
						m_fOldTime = m_fTotalFPS = 0.0f;
						// First segment						
						m_curInterpolation.fTime -= m_arPath->pPrevious->fTime;						
					}	
					// Recompute factor
					fDif = m_curSeg->pNext->fTime - m_curSeg->fTime;		                
					s = (m_curInterpolation.fTime - m_curSeg->fTime) / abs(fDif);					
				}
				// Previous segment or last segment
				if((s < 0.0f || m_curSeg == m_arPath->pPrevious) && m_fSpeed < 0.0f)
				{
					// Previous segment
					if(m_curSeg->pPrevious)
					{
						m_curSeg = m_curSeg->pPrevious;						
					}
					// Set max time for last segment
					if(m_curSeg == m_arPath->pPrevious)
					{					
						m_curInterpolation.fTime += m_arPath->pPrevious->fTime;						
					}	
					// Recompute factor
					fDif = m_curSeg->pNext->fTime - m_curSeg->fTime;		                
					s = (m_curInterpolation.fTime - m_curSeg->fTime) / abs(fDif);	
				}
				ps1 = ps2 = ps3 = ps4 = m_curSeg;
				if(m_curSeg->pPrevious)
				{
					ps1 = m_curSeg->pPrevious;
				}				
				if(m_curSeg->pNext)
				{					
					ps3 = m_curSeg->pNext;
				}
				if(m_curSeg->pNext->pNext)
				{					
					ps4 = m_curSeg->pNext->pNext;
				}
	
				if(m_dwFlags & CF_CATMULLROMINT)
				{
					D3DXVec3CatmullRom(&m_curInterpolation.vPos, &ps1->vPos, &ps2->vPos, &ps3->vPos, &ps4->vPos, s);
					D3DXVec3CatmullRom(&m_curInterpolation.vDir, &ps1->vDir, &ps2->vDir, &ps3->vDir, &ps4->vDir, s);
				}
				else
				{
					D3DXVec3Lerp(&m_curInterpolation.vDir, &ps2->vDir, &ps3->vDir, s);
					D3DXVec3Lerp(&m_curInterpolation.vPos, &ps2->vPos, &ps3->vPos, s);
				}

				m_curInterpolation.fTime += fElapsedTime;	

				if(m_curInterpolation.fTime - m_fOldTime > 1)
				{
					CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
					if(!dis)
					{
						XException exp("LoadTextureFromFile: Device is invalid.");
						THROW(exp);			
					}
					m_fTotalFPS += dis->GetCurrentFPS();
					m_fOldTime = m_curInterpolation.fTime;
					SAFE_RELEASE(dis);
				}
			}
			m_vEyePt = m_curInterpolation.vPos;
			m_vLookAtPoint = m_vEyePt + m_curInterpolation.vDir;
		}
		//
		// Create new view matrix
		D3DXMatrixLookAtLH(&m_matView, &m_vEyePt, &m_vLookAtPoint, &vUpDir);
		// Apply transfromation		
		m_lpDevice->SetTransform(D3DTS_VIEW, &m_matView);
		//
		// Reset velocities
		m_fScrlVelX = 0;
		m_fScrlVelY = 0;
		m_fRotVel = 0;
		m_fZoomVel = 0;
		m_fZAngleVel = 0;
				
		dwRet = ERROR_SUCCESS;
		//
		// Computes billboard matrix
		D3DXVECTOR3 vDir = m_vLookAtPoint - m_vEyePt;
		if( vDir.y > 0.0f ) {
			D3DXMatrixRotationZ(&m_matBillboard, -atanf(vDir.x/vDir.y)+D3DX_PI/2 );
		}
		else {
			D3DXMatrixRotationZ(&m_matBillboard, -atanf(vDir.x/vDir.y)-D3DX_PI/2 );
		}			

		//
		// Compute clip volume
		ComputeClipVolume();

		if(m_dwFlags & CF_DRAWINFO)
		{
			static char szCameraInfo[128];
			if(m_dwFlags & CF_PATH_MODE)
			{
				sprintf(szCameraInfo, "LookAtPoint: %2.1lf, %2.1lf, %2.1lf\nEyePoint: %2.1lf, %2.1lf, %2.1lf\nMode: %s\nAverage FPS: %2.2lf - Segment: %2.1lf/%2.1lf - Method: %s", m_vLookAtPoint.x, m_vLookAtPoint.y, m_vLookAtPoint.z, m_vEyePt.x, m_vEyePt.y, m_vEyePt.z, (m_dwFlags & CF_PATH_MODE) ? "Path" : "Interactive", m_fAvgFPS, m_curInterpolation.fTime, m_arPath->pPrevious->fTime, (m_dwFlags & CF_CATMULLROMINT) ? "Catmull-Rom" : "Linear");		
			}
			else
			{
				sprintf(szCameraInfo, "LookAtPoint: %2.1lf, %2.1lf, %2.1lf\nEyePoint: %2.1lf, %2.1lf, %2.1lf\nMode: %s", m_vLookAtPoint.x, m_vLookAtPoint.y, m_vLookAtPoint.z, m_vEyePt.x, m_vEyePt.y, m_vEyePt.z, (m_dwFlags & CF_PATH_MODE) ? "Path" : "Interactive");		
			}
			m_lpCameraInfo->DrawText(szCameraInfo, 0, 50, D3DCOLOR_ARGB(255,220,220,255));		
		}
	}
	return dwRet;

}

void XCamera::ComputeClipVolume()
{
	FLOAT dist, t1, t2;
	D3DXVECTOR3 p, pt[8];
	D3DXVECTOR3 v1, v2, n;
	D3DXVECTOR3 vUpDir = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
	D3DXVECTOR3 vDir = m_vLookAtPoint - m_vEyePt ;
	D3DXVECTOR3 vCross;
	D3DXVec3Cross(&vCross, &vUpDir, &vDir);
	D3DXVec3Cross(&vUpDir, &vDir, &vCross);
	D3DXVec3Normalize(&vDir, &vDir);
	D3DXVec3Normalize(&vCross, &vCross);
	D3DXVec3Normalize(&vUpDir, &vUpDir);

	/*  
	*---------------------------------------*
	|					^vUp * t1			|
	|					|					|
	|					^vUp				|
	|					|					|
	|<---------------<--*Cam				|
	|vCross * t2   vCross  					|
	|										|
	|										|
	|										|
	*---------------------------------------*
						  /
						/
					  /
					/
				  /
		 dist	/
	|<--------/>^vUp * t1
	|		/   |
	|	  /     |
	^	/       |t1
 vUp| /         |
	|           |
	Cam->vDir-->|dist * vDir
	\   FOV     |
	  \         |
	    \       |-t1
		  \     |
			\   |
			  \ |
				\
				  \
					\
					  \
						\
						  \
	
	*/
	for(INT i = 0; i < 8; i++)
	{
		dist = (i & 0x4) ? FAR_CLIP : NEAR_CLIP;
		pt[i] = dist * vDir;
		t1 = dist * tanf(FOV/2);
		t1 = (i & 0x2) ? t1 : -t1;
		pt[i] += vUpDir * t1;
		t2 = dist * tanf(FOV/2) * ASPECT; // take into account screen proportions
		t2 = (i & 0x1) ? -t2 : t2;
		pt[i] += vCross * t2;
		pt[i] += m_vEyePt;
	}

	//compute the near plane
	v1 = pt[2] - pt[0];
	v2 = pt[1] - pt[0];
	D3DXVec3Cross(&n, &v2, &v1);
	D3DXVec3Normalize(&n, &n);	
	m_CV.pNear.a = n.x;
	m_CV.pNear.b = n.y;
	m_CV.pNear.c = n.z;
	m_CV.pNear.d = -(n.x * pt[0].x + n.y * pt[0].y + n.z * pt[0].z);

	//compute the far plane
	v1 = pt[5] - pt[4];
	v2 = pt[6] - pt[4];
	D3DXVec3Cross(&n, &v2, &v1);
	D3DXVec3Normalize(&n, &n);
	m_CV.pFar.a = n.x;
	m_CV.pFar.b = n.y;
	m_CV.pFar.c = n.z;
	m_CV.pFar.d = -(n.x * pt[4].x + n.y * pt[4].y + n.z * pt[4].z);// - 10.0f;

	//compute the top plane
	v1 = pt[6] - pt[2];
	v2 = pt[3] - pt[2];
	D3DXVec3Cross(&n, &v2, &v1);
	D3DXVec3Normalize(&n, &n);
	m_CV.pTop.a = n.x;
	m_CV.pTop.b = n.y;
	m_CV.pTop.c = n.z;
	m_CV.pTop.d = -(n.x * pt[2].x + n.y * pt[2].y + n.z * pt[2].z);

	//compute the bottom plane
	v1 = pt[1] - pt[0];
	v2 = pt[4] - pt[0];
	D3DXVec3Cross(&n, &v2, &v1);
	D3DXVec3Normalize(&n, &n);
	m_CV.pBottom.a = n.x;
	m_CV.pBottom.b = n.y;
	m_CV.pBottom.c = n.z;
	m_CV.pBottom.d = -(n.x * pt[0].x + n.y * pt[0].y + n.z * pt[0].z);

	//compute the left plane
	v1 = pt[3] - pt[1];
	v2 = pt[5] - pt[1];
	D3DXVec3Cross(&n, &v2, &v1);
	D3DXVec3Normalize(&n, &n);
	m_CV.pLeft.a = n.x;
	m_CV.pLeft.b = n.y;
	m_CV.pLeft.c = n.z;
	m_CV.pLeft.d = -(n.x * pt[1].x + n.y * pt[1].y + n.z * pt[1].z);

	//compute the right plane
	v1 = pt[4] - pt[0];
	v2 = pt[2] - pt[0];
	D3DXVec3Cross(&n, &v2, &v1);
	D3DXVec3Normalize(&n, &n);
	m_CV.pRight.a = n.x;
	m_CV.pRight.b = n.y;
	m_CV.pRight.c = n.z;
	m_CV.pRight.d = -(n.x * pt[0].x + n.y * pt[0].y + n.z * pt[0].z);
}

HRESULT	XCamera::LoadPath(LPCSTR szPath)
{
	DWORD size;
	char line[256];
	char *pline = line;
	char c;		
	BOOL bLookAtPoint = FALSE;
	PathSegment *pLast = NULL;
	PathSegment ps;

	char szFullPath[MAX_PATH];
	// Build full path to path file
	cmnGetDataFilePath(szFullPath, szPath);
	// Load file contains
	cmnLoadFileFromPath(szFullPath, NULL, &size);
	char * file = new char[size+1];
	ZeroMemory(file, (size+1) * sizeof(char));
	cmnLoadFileFromPath(szFullPath, (LPBYTE)file, &size);
	
	// Release previous path if any
	if(m_arPath)
	{	
		PathSegment *pNext = NULL;
		PathSegment *pToDelete = m_arPath;
		// Release path
		while(pToDelete)
		{
			pNext = pToDelete->pNext;
			SAFE_DELETE(pToDelete);
			pToDelete = pNext;
		}		
	}
	char *pfile = file;	
	m_arPath = NULL;

	while(pfile[0] != '\0')
	{
		while(1)
		{
			c = pfile++[0];
			if(c == 0x0D)
			{
				pfile++; // jump over 0x0A
				break;
			}
			if(c == '\0')
			{
				pfile--; // return before '\0'
				break;
			}
			// Copy line char
			pline[0] = c;
			// Next char
			pline++;			
		}
		// End of string
		pline[0] = '\0';
		// Reset pline
		pline = line;

		// Second vector is look at point
		if(pline[0] == '#')
		{
			bLookAtPoint = TRUE;
			// Jump over '#'
			pline++;
		}
		if(pline[0] == '$')
		{
			// Jump over '$'
			pline++;
		}
		// Jump over comment
		if(pline[0] == '!')
		{
			continue;
		}
		// Parse line
		ps.fTime = ExtractNumber(0, pline);
		ps.vPos.x = ExtractNumber(1, pline);
		ps.vPos.y = ExtractNumber(2, pline);
		ps.vPos.z = ExtractNumber(3, pline);
		ps.vDir.x = ExtractNumber(4, pline);
		ps.vDir.y = ExtractNumber(5, pline);
		ps.vDir.z = ExtractNumber(6, pline);

		// vDir is look at point
		if(bLookAtPoint)
		{
			ps.vDir = ps.vDir - ps.vPos;
			bLookAtPoint = FALSE;
		}
		// Normalize direction
		D3DXVec3Normalize(&ps.vDir, &ps.vDir);
		
		if(!m_arPath)
		{
			// Create root segment
			pLast = m_arPath = new PathSegment;
			*m_arPath = ps;
			m_arPath->pPrevious = NULL;
			m_arPath->pNext = NULL;			
		}
		else
		{
			// Create new segment			
			pLast->pNext = new PathSegment;			
			*pLast->pNext = ps; 
			pLast->pNext->pPrevious = pLast;
			pLast->pNext->pNext = NULL;
			pLast = pLast->pNext;
		}				

		// Reset pline
		pline = line;
	}	
	// Cycle path - last next is first and first previous is last
	pLast->pNext = m_arPath;
	m_arPath->pPrevious = pLast;
		
	SAFE_DELETE_ARRAY(file);

	// Path is ready
	m_bPathLoaded = TRUE;	
	m_curInterpolation = *m_arPath;
	m_curSeg = m_arPath;
	m_vEyePt = m_curInterpolation.vPos;
	m_vLookAtPoint = m_vEyePt + m_curInterpolation.vDir;
	
	return S_OK;
}

float	XCamera::ExtractNumber(int i, char *line)
{
	//float ret;
	char number[20];
	char *pline = line;
	int c;

	for(int u = 0; u < i; u++)
	{
		// Find first ;
		c = strcspn(pline, ";");		
		// Next number + ;
		if(c > 0)
		{
			pline += (c+1); 
		}
	}

	// Find first ;
	c = strcspn(pline, ";");
	// Copy number
	strncpy(number, pline, c);
	// Ends number
	number[c] = '\0';
	// Convert number
	return (float)atof(number);
}

HRESULT XCamera::StartPath()
{
	if(m_bPathLoaded)
	{
		m_ePathStatus = PLAY;
		m_dwFlags |= CF_PATH_MODE;
		return S_OK;
	}
	else
	{
		return S_FALSE;
	}
}

HRESULT XCamera::PausePath()
{
	if(m_bPathLoaded)
	{
		m_ePathStatus = PAUSE;
		return S_OK;
	}
	else
	{
		return S_FALSE;
	}	
}

HRESULT XCamera::StopPath()
{
	if(m_bPathLoaded)
	{
		m_ePathStatus = STOP;
		// Reset path
		m_curInterpolation = *m_arPath;
		m_curSeg = m_arPath;
		return S_OK;
	}
	else
	{
		return S_FALSE;
	}	
}

