// XCamera.h: interface for the XCamera class.
//
//////////////////////////////////////////////////////////////////////

#pragma once

#define		_C_MAX_ROTATION_SPEED		5.0f
#define		_C_MAX_ZOOMING_SPEED		20.0f
#define		_C_MAX_SCROLLING_SPEED		20.0f
#define		_C_MAX_ZANGLE_SPEED			2.0f

#define		MAX_ZOOM					300.0f
#define		MIN_ZOOM					5.0f

#define		Z_ANGLE						D3DX_PI / 2

#define		MAX_Z_ANGLE					D3DX_PI / 2
#define		MIN_Z_ANGLE					D3DX_PI / 7 

const float NEAR_CLIP = 1.0f;
const float FAR_CLIP = 1000.0f;
const float FOV = D3DX_PI / 4.0f;
const float ASPECT = 4.0f / 3.0f;

#define		CF_LINEARINT				0x0000
#define		CF_CATMULLROMINT			0x0001
#define		CF_DRAWINFO					0x0002
#define		CF_PATH_MODE				0x0004

#include "Defs.h"

struct PathSegment
{
	float			fTime;
	D3DXVECTOR3		vPos;
	D3DXVECTOR3		vDir;

	PathSegment *pNext;
	PathSegment *pPrevious;
};

class XCamera : public ICamera
{
	// D3D device object...
	LPDIRECT3DDEVICE8	m_lpDevice;
	I3DFont				*m_lpCameraInfo;
	//
	// Camera bounds
	int					m_iMinPos;
	int					m_iMaxPos; 
	//
	// Velocities in all directions
	float				m_fScrlVelX;	
	float				m_fScrlVelY;	
	float				m_fRotVel;
	float				m_fZoomVel;
	float				m_fZAngleVel;
	//
	// Camera position
	D3DXVECTOR3			m_vLookAtPoint;
	D3DXVECTOR3			m_vEyePt;
	float				m_fZoomDistance;
	float				m_fZAngle;		
	float				m_fViewAngle;
	float				m_fCameraHeight;
	//
	// Camera matrices
	D3DXMATRIX			m_matBillboard;		
	D3DXMATRIX			m_matProjection;
	D3DXMATRIX			m_matView;	

	//
	// Current clip volume
	CLIPVOLUME			m_CV;

	DWORD				m_dwRef;

	//
	// Computes clip volume each cycle
	void				ComputeClipVolume();

	// Cela path
	PathSegment *m_arPath;
	// Current position of the camera
	PathSegment  *m_curSeg;
	PathSegment  m_curInterpolation;
	// FALSE = stop, TRUE = started	
	PATHSTATUS  m_ePathStatus;
	// Motion modifier
	float		m_fSpeed;
	float		m_fAvgFPS;

	// Path loaded successfully
	BOOL	m_bPathLoaded;

	// Benchmark
	float		m_fTotalFPS;
	float		m_fOldTime;

	DWORD		m_dwFlags;

public:
	//
	// Inicialization
	virtual HRESULT				InitCamera(int iMinBound, int iMaxBound);
	// Update
	virtual HRESULT				ProcessCamera(float	fElapsedTime);	
	// Call after device reset
	virtual	HRESULT				RestoreCamera();
	// Camera motion
	virtual void				ZoomIn() {m_fZoomVel = -(_C_MAX_ZOOMING_SPEED);}
	virtual void				ZoomOut() {m_fZoomVel = _C_MAX_ZOOMING_SPEED;}
	virtual void				RotateLeft() {m_fRotVel = _C_MAX_ROTATION_SPEED;}
	virtual void				RotateRight() {m_fRotVel = -(_C_MAX_ROTATION_SPEED);}
	virtual void				MoveLeft() {m_fScrlVelX = _C_MAX_SCROLLING_SPEED;}
	virtual void				MoveRight() {m_fScrlVelX = -(_C_MAX_SCROLLING_SPEED);}
	virtual void				MoveUp() {m_fScrlVelY = -(_C_MAX_SCROLLING_SPEED);}
	virtual void				MoveDown() {m_fScrlVelY = _C_MAX_SCROLLING_SPEED;}
	virtual void				IncreaseZAngle() {m_fZAngleVel =   _C_MAX_ZANGLE_SPEED;}
	virtual void				DecreaseZAngle() {m_fZAngleVel = -(_C_MAX_ZANGLE_SPEED);}
	virtual void				SetCameraHeight(float fHeight) {m_fCameraHeight = fHeight;}			
	//
	// Get methods
	virtual float				GetViewAngle() {return m_fViewAngle;}
	virtual float				GetZoom() {	return m_fZoomDistance;}
	virtual float				GetZAngle() {return m_fZAngle;}
	virtual D3DXVECTOR3*		GetCameraPos() {return &m_vLookAtPoint;}
	virtual D3DXVECTOR3*		GetEyePoint() {return &m_vEyePt;}
	virtual D3DXMATRIX*			GetBillboardMatrix() {return &m_matBillboard;}
	virtual void				GetWorldViewProjMatrix(D3DXMATRIX * matWVP) {;}
	virtual void				GetClipVolume(CLIPVOLUME & cv) {cv = m_CV;}
	
	virtual HRESULT		AddRef();
	virtual HRESULT		Release();

	// Path methods
	// Path file:
	// time;px;py;pz;dx;dy;dz;
	// px-z are position vector
	// dx-z are direction vector
	// nexttime; next position vector; next direction vector;
	virtual HRESULT	LoadPath(LPCSTR szPath);
	virtual HRESULT StartPath();
	virtual HRESULT PausePath();
	virtual HRESULT StopPath();// reset
	virtual HRESULT SetSpeed(float fRelativeSpeed) {m_fSpeed += fRelativeSpeed;return S_OK;}

	float	ExtractNumber(int i, char *line);

	void		SetFlag(DWORD dwFlag) {m_dwFlags |= dwFlag;}
	void		UnsetFlag(DWORD dwFlag) {m_dwFlags &= ~dwFlag;}
	void		InvertFlag(DWORD dwFlag) { if(dwFlag & m_dwFlags) { UnsetFlag(dwFlag); } else { SetFlag(dwFlag);}}
	void		ResetFlags() {m_dwFlags = 0;}


	XCamera();
	~XCamera();
};

